/*************************************************************************
 *  midi.c
 *
 *  Copyright (C) 1995 Michael Heasley (mheasley@hmc.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "musserver.h"


struct synth_info sinfo[5];
struct midi_info minfo[5];
int seqfd;
int use_midi = 0;
int use_synth = 0;
int seq_dev;
struct opl_instr fm_instruments[150];
extern int verbose;

SEQ_DEFINEBUF(2048);

void seqbuf_dump(void)
{
  if (_seqbufptr)
  if (write(seqfd, _seqbuf, _seqbufptr) == -1)
    {
     perror("write /dev/sequencer");
     exit (-1);
     }
  _seqbufptr = 0;
}

void midi_setup(int pref_dev)
{
int numsynths;
int m = 0;
int s = 0;
int have_synth = 1;
int have_midi = 1;

  if (pref_dev != EXT_MIDI)
    if (ioctl(seqfd, SNDCTL_SEQ_NRSYNTHS, &numsynths) == -1)
      if (pref_dev == FM_SYNTH)
        {
        printf("No synth devices found.  Exiting.\n");
        exit(1);
        }
      else
        have_synth = 0;
    else
      for (s = 0; s < numsynths; s++)
        {
        sinfo[s].device = s;
        ioctl(seqfd, SNDCTL_SYNTH_INFO, &sinfo[s]);
        if (verbose)
          printf("Found a synth device: type %d (%s)\n", sinfo[s].synth_type, sinfo[s].name);
        }
  else
    have_synth = 0;

  if (pref_dev != FM_SYNTH)
    if (ioctl(seqfd, SNDCTL_SEQ_NRMIDIS, &numsynths) == -1)
      {
      if (pref_dev == EXT_MIDI)
        {
        printf("No midi devices found.  Exiting.\n");
        exit(1);
        }
      else if (have_synth == 0)
        {
        printf("No synth or midi devices found.  Exiting.\n");
        exit(1);
        }
      else
        have_midi = 0;
      }
    else
      for (m = 0; m < numsynths; m++)
        {
        minfo[m].device = m;
        ioctl(seqfd, SNDCTL_MIDI_INFO, &minfo[m]);
        if (verbose)
          printf("Found a midi device: type %d (%s)\n", minfo[m].dev_type, minfo[m].name);
        }
  else
    have_midi = 0;

  if (have_midi == 1)
    {
    use_midi = 1;
    seq_dev = m - 1;
    if (verbose)
      printf("Using midi device number %d (%s)\n", m, minfo[m - 1].name);
    }
  else
    {
    use_synth = 1;
    seq_dev = s - 1;
    if (verbose)
      printf("Using synth device number %d (%s)\n", s, sinfo[s - 1].name);
    }
}

void reset_midi(void)
{
unsigned int channel;

  if (use_midi)
    for (channel = 0; channel < 16; channel++)
      {
      /* all notes off */
      SEQ_MIDIOUT(seq_dev, MIDI_CTL_CHANGE + channel);
      SEQ_MIDIOUT(seq_dev, 0x7b);
      SEQ_MIDIOUT(seq_dev, 0);
      /* reset pitch bender */
      SEQ_MIDIOUT(seq_dev, MIDI_PITCH_BEND + channel);
      SEQ_MIDIOUT(seq_dev, 64 >> 7);
      SEQ_MIDIOUT(seq_dev, 64 & 127);
      /* reset volume to 100 */
      SEQ_MIDIOUT(seq_dev, MIDI_CTL_CHANGE + channel);
      SEQ_MIDIOUT(seq_dev, CTL_MAIN_VOLUME);
      SEQ_MIDIOUT(seq_dev, 100);
      /* reset pan */
      SEQ_MIDIOUT(seq_dev, MIDI_CTL_CHANGE + channel);
      SEQ_MIDIOUT(seq_dev, CTL_PAN);
      SEQ_MIDIOUT(seq_dev, 64);

      SEQ_DUMPBUF();
      }
  else
    for (channel = 0; channel < sinfo[seq_dev].nr_voices; channel++)
      {
      int note;
      for (note = 0; note < 128; note++)
        SEQ_STOP_NOTE(seq_dev, channel, note, 64);
      }
}

void note_off(int note, int channel, int volume)
{
  if (use_midi)
    {
    SEQ_MIDIOUT(seq_dev, MIDI_NOTEOFF + channel);
    SEQ_MIDIOUT(seq_dev, note);
    SEQ_MIDIOUT(seq_dev, volume);
    SEQ_DUMPBUF();
    }
  else
    SEQ_STOP_NOTE(seq_dev, channel, note, volume);
}

void note_on(int note, int channel, int volume)
{
  if (use_midi)
    {
    SEQ_MIDIOUT(seq_dev, MIDI_NOTEON + channel);
    SEQ_MIDIOUT(seq_dev, note);
    SEQ_MIDIOUT(seq_dev, volume);
    SEQ_DUMPBUF();
    }
  else
    SEQ_START_NOTE(seq_dev, channel, note, volume);
}

void pitch_bend(int channel, int value)
{
  if (use_midi)
    {
    SEQ_MIDIOUT(seq_dev, MIDI_PITCH_BEND + channel);
    SEQ_MIDIOUT(seq_dev, value >> 7);
    SEQ_MIDIOUT(seq_dev, value & 127);
    SEQ_DUMPBUF();
    }
  else
    SEQ_BENDER(seq_dev, channel, value);
}

void control_change(int controller, int channel, int value)
{
  if (use_midi)
    {
    SEQ_MIDIOUT(seq_dev, MIDI_CTL_CHANGE + channel);
    SEQ_MIDIOUT(seq_dev, controller);
    SEQ_MIDIOUT(seq_dev, value);
    SEQ_DUMPBUF();
    }
  else
    SEQ_CONTROL(seq_dev, channel, controller, value);
}

void patch_change(int patch, int channel)
{
  if (use_midi)
    {
    SEQ_MIDIOUT(seq_dev, MIDI_PGM_CHANGE + channel);
    SEQ_MIDIOUT(seq_dev, patch);
    SEQ_DUMPBUF();
    }
  else
    {
    struct sbi_instrument data;
    data.key = FM_PATCH;
    data.device = seq_dev;
    data.channel = channel;
    strcpy(data.operators, fm_instruments[patch].patchdata);
/*    ioctl(seqfd, SNDCTL_FM_LOAD_INSTR, data); */
    SEQ_WRPATCH(&data, sizeof(data));
    }
}

void midi_wait(float time)
{
  SEQ_WAIT_TIME((int) time);
  SEQ_DUMPBUF();
  ioctl(seqfd, SNDCTL_SEQ_SYNC);
}

void midi_timer(int action)
{
  switch (action)
    {
    case 0:
      SEQ_START_TIMER();
      break;
    case 1:
      SEQ_STOP_TIMER();
      break;
    case 2:
      SEQ_CONTINUE_TIMER();
      break;
    }
}
