/*************************************************************************
 *  musserver.c
 *
 *  Copyright (C) 1995 Michael Heasley (mheasley@hmc.edu)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include "musserver.h"

extern int seqfd;
extern int use_synth;
unsigned long lumpsize[35];	/* size in bytes of data lumps */
unsigned long lumppos[35];	/* position of lumps in wad file */
unsigned char *musicdata;
int lump = 0;
int level = 1;
int doomver;
int inter = 0;
int inter2 = 32;
int index = 0;
int skip = 0;
int verbose = 0;

void readdir(FILE *wadfile, int version);
int readmus(FILE *wadfile, int lumpnum);
int playmus(char *musdata, unsigned int mussize, int play_once);
void midi_setup(int pref_dev);
void read_genmidi(FILE *wadfile);

static int musicmap[3][76] = {
                         {27, 0, 29, 1, 29, 2, 29, 3, 29, 4, 29, 5, 29, 6,
                          29, 7, 30, 100, 205, 8, 29, 207, 100, 100, 100, 100,
                          100, 100, 100, 100, 100, 100, 100, 100, 100, 100,
                          100, 100, 100, 100, 100, 100, 100, 100, 100, 100,
                          100, 100, 100, 100, 100, 100, 100, 100, 100, 100,
                          100, 100, 100, 100, 100, 100, 100, 100, 100, 100,
                          100, 100, 100, 100, 100, 100, 100, 100, 100, 100},
                         {27, 0, 29, 1, 29, 2, 29, 3, 29, 4, 29, 5, 29, 6,
                          29, 7, 30, 100, 205, 8, 29, 207, 9, 29, 10, 29, 11,
                          29, 12, 29, 13, 29, 14, 29, 15, 29, 16, 30, 100, 230,
                          17, 29, 232, 18, 29, 19, 29, 20, 29, 21, 29, 22, 29,
                          23, 29, 24, 29, 25, 30, 31, 100, 253, 26, 29, 255,
                          100, 100, 100, 100, 100, 100, 100, 100, 100, 100,
                          100},
                         {33, 0, 34, 1, 34, 2, 34, 3, 34, 4, 34, 5, 34, 32, 6,
                           34, 7, 34, 8, 34, 9, 34, 10, 34, 32, 11, 34, 12, 34,
                           13, 34, 14, 34, 15, 34, 16, 34, 17, 34, 18, 34, 19,
                           34, 32, 20, 34, 21, 34, 22, 34, 23, 34, 24, 34, 25,
                           34, 26, 34, 27, 34, 28, 34, 29, 34, 32, 30, 100,
                           231, 32, 30, 34, 233, 32, 31, 34, 233} };

static int secretmap[3][3] = {
                           {19, 100, 100},
                           {19, 40, 62},
                           {68, 72, 100} };

void show_help(void)
{
  printf("Usage: musserver [options]\n");
  printf("\n  The following options cause musserver to send a signal to an existing\n");
  printf("  musserver process and then exit:\n");
  printf("  -n               Go on to next music lump\n");
  printf("  -p               Go back to previous level's music\n");
  printf("  -s               Go on to secret level\n");
  printf("  -t               Terminate existing musserver process\n");
  printf("\n  The following options apply when starting a new musserver process:\n");
  printf("  -d directory     Look in 'directory' for DOOM wad file\n");
  printf("  -f               Use FM synth device for music playback\n");
  printf("  -h               Print this message and exit\n");
  printf("  -i               Do not play intermission music between levels\n");
  printf("  -k               Kill existing musserver process and start a new one\n");
  printf("  -m               Use general midi device for music playback\n");
  printf("  -v               Verbose\n");
  printf("  -w level         Warp to 'level'\n");
  printf("      In DOOM I:   The first digit of 'level' is episode number\n");
  printf("                   The second digit of 'level' is mission number\n");
  printf("                   i.e. 31 means E3M1\n");
  printf("      In DOOM II:  'level' is the level number to warp to\n");
}
 
void kill_muspid(int sig)
{
FILE *pidfile;
int pid;

  pidfile = fopen("/tmp/musserver.pid", "r");
  if (pidfile == NULL)
    switch(sig)
      {
      case SIGKILL:
        break;
      default:
        printf("Error: could not find a running copy of musserver (missing file\n/tmp/musserver.pid)\n");
        exit(1);
        break;
      }
  else
    {
    fscanf(pidfile, "%d", &pid);
    fclose(pidfile);
    kill(pid, sig);
    }
}

void next(void)
{
  index++;
  if (skip)
    while ((musicmap[doomver][index] == inter) || (musicmap[doomver][index] == inter2))
      index++;
  lump = musicmap[doomver][index];
  if ((lump != inter) && (lump != inter2))
    level = lump + 1;
}

void prev(void)
{
  if ((index > 1) && ((doomver == 2) || ((lump != 9) && (lump != 18))))
    {
    index--;
    while ((musicmap[doomver][index] == inter) || (musicmap[doomver][index] == inter2) || (index == 71))
      index--;
    lump = musicmap[doomver][index];
    level = lump + 1;
    }
}

void secret(void)
{
int episode;

  if (doomver < 2)
    {
    episode = ((level - 1) / 9);
    index = secretmap[doomver][episode];
    }
  else
    {
    if (index > 66)
      index = secretmap[doomver][1];
    else
      index = secretmap[doomver][0];
    if (skip)
      index++;
    }
  lump = musicmap[doomver][index];
  level = lump + 1;
}

void main(int argc, char **argv)
{
FILE *infile;
FILE *outfile;
int done = 0;
int result;
char *wadfilename;
char *waddir;
unsigned int musicsize;
char x;
int intro = 0;
int outtro = 31;
int opt_dev = 0;
int warp = 0;
int episode;
int mission;


  waddir = getenv("DOOMWADDIR");
  if (waddir == NULL)
    {
    waddir = malloc(1);
    sprintf(waddir,  ".");
    }
  while ((x = getopt(argc, argv, "d:fhikmnpstvw:")) != -1)
    switch (x)
      {
      case 'd':
        waddir = optarg;
        break;
      case 'f':
        opt_dev = FM_SYNTH;
        break;
      case 'h':
        show_help();
        exit(0);
        break;
      case 'i':
        skip = 1;
        break;
      case 'k':
        kill_muspid(SIGKILL);
        unlink("/tmp/musserver.pid");
        break;
      case 'm':
        opt_dev = EXT_MIDI;
        break;
      case 'n':
        kill_muspid(SIGHUP);
        exit(0);
        break;
      case 'p':
        kill_muspid(SIGUSR1);
        exit(0);
        break;
      case 's':
        kill_muspid(SIGUSR2);
        exit(0);
        break;
      case 't':
        kill_muspid(SIGTERM);
        exit(0);
        break;
      case 'v':
        verbose = 1;
        break;
      case 'w':
        warp = atoi(optarg);
        break;
      case '?': case ':':
        show_help();
        exit(1);
        break;
      }

  outfile = fopen("/tmp/musserver.pid", "r");
  if (outfile != NULL)
    {
    printf("Another copy of musserver appears to be already running.  Use the -k option\n");
    printf("to kill that copy and start a new one.\n");
    fclose(outfile);
    exit(1);
    }

  wadfilename = malloc(strlen(waddir) + strlen("/doom1.wad") + 1);
  sprintf(wadfilename, "%s/doom1.wad", waddir);
  infile = fopen(wadfilename, "r");
  if (infile == NULL)
    {
    wadfilename = malloc(strlen(waddir) + strlen("/doom.wad") + 1);
    sprintf(wadfilename, "%s/doom.wad", waddir);
    infile = fopen(wadfilename, "r");
    if (infile == NULL)
      {
      wadfilename = malloc(strlen(waddir) + strlen("/doom2.wad") + 1);
      sprintf(wadfilename, "%s/doom2.wad", waddir);
      infile = fopen(wadfilename, "r");
      if (infile == NULL)
        {
        printf("Game mode indeterminate.\n");
        exit(1);
        }
      else
        doomver = 2;
      }
    else
      doomver = 1;
    }
  else
    doomver = 0;

  if (verbose)
    switch (doomver)
      {
      case 0:
        printf("Playing music for shareware DOOM\n");
        break;
      case 1:
        printf("Playing music for registered DOOM\n");
        break;
      case 2:
        printf("Playing music for DOOM II\n");
        break;
      }

  switch(doomver)
    {
    case 0: case 1:
      inter = 29;
      intro = 27;
      if (warp)
        {
        level = (((warp / 10) - 1) * 9) + (warp % 10);
        lump = level - 1;
        if ((level < 1) || (level > 27))
          {
          printf("Illegal level number: %d\n", warp);
          exit(1);
          }
        }
      break;
    case 2:
      inter = 34;
      intro = 33;
      outtro = 100;
      if (warp)
        {
        level = warp;
        lump = level - 1;
        if ((level < 1) || (level > 32))
          {
          printf("Illegal level number: %d", warp);
          exit(1);
          }
        }
      break;
    }
  if (!lump)
    lump = intro;
  else
    {
    for (x = 0; x < 72; x++)
      if (musicmap[doomver][(int)x] == lump)
        {
        index = x;
        break;
        }
    if (x == 72)
      {
      printf("Error: couldn't find position of level %d's music.  Exiting.\n", warp);
      exit(1);
      }
    }

/* write pid to tempfile */
  outfile = fopen("/tmp/musserver.pid", "w");
  if (outfile == NULL)
    {
    printf("Could not open temp file /tmp/musserver.pid for writing!\n");
    printf("You may experience problems later on.\n");
    }
  else
    {
    fprintf(outfile, "%d\n", getpid());
    fclose(outfile);
    }

/* read the wadfile, get music data */
  readdir(infile, doomver);

  seqfd = open("/dev/sequencer", O_WRONLY, 0);
  midi_setup(opt_dev);
  if (use_synth)
    read_genmidi(infile);
  while (!done)
    {
/* neither condition (lump == 100 or lump > 200) should occur here, but just
   to be on the safe side . . . */
    if (lump == 100)
      break;
    else if (lump > 200)
      {
      index = lump - 200;
      lump = musicmap[doomver][index];
      level = lump + 1;
      }
    if (verbose)
      switch (doomver)
        {
        case 0: case 1:
          episode = ((level - 1) / 9) + 1;
          mission = ((level - 1) % 9) + 1;
          switch (lump)
            {
            case 27:
              printf("Playing intro\n");
              break;
            case 29:
              printf("Playing intermission\n");
              break;
            case 30:
              printf("Playing victory\n");
              break;
            case 31:
              printf("Playing bunny ending\n");
              break;
            default:
              printf("Playing level E%dM%d\n", episode, mission);
              break;
            }
          break;
        case 2:
          switch(lump)
            {
            case 33:
              printf("Playing intro\n");
              break;
            case 34:
              printf("Playing intermission\n");
              break;
            case 32:
              printf("Playing special intermission\n");
              break;
            case 30:
              if (index == 65)
                printf("Playing ending\n");
              else
                printf("Playing level %d\n", level);
              break;
            default:
              printf("Playing level %d\n", level);
              break;
            }
          break;
        }
           
    musicsize = readmus(infile, lump);
    result = playmus(musicdata, musicsize, ((lump == intro) || (lump == outtro)) ? 1 : 0);
    switch (result)
      {
      case -1:
        prev();
        break;
      case 1:
        next();
        break;
      case 2:
        secret();
        break;
      case 3:
        done = 1;
        break;
      default:
        done = 1;
        printf("Error in music playing.\n");
        break;
      }
    if (lump == 100)
      done = 1;
    else if (lump > 200)
      {
      index = lump - 200;
      lump = musicmap[doomver][index];
      level = lump + 1;
      }
    }
  fclose(infile);
  unlink("/tmp/musserver.pid");
}
